<?php

namespace Concrete\Package\JlPackageMagicPrb;

use Package;
use Events;
use Concrete\Core\Support\Facade\Application;
use Illuminate\Filesystem\Filesystem;

defined('C5_EXECUTE') or die("Access Denied.");

/*
 * Package Magic PRB by John Liddiard (aka JohntheFish)
 * www.c5magic.co.uk
 *
 * This package contains software copyright and proprietary to John Liddiard
 *
 */

class Controller extends Package {

    protected $pkgHandle = 'jl_package_magic_prb';
    protected $appVersionRequired = '5.7.5.8';
    protected $pkgVersion = '1.0.3';

    public function getPackageDescription() {
        return t('Integrate a PRB source for packages to Package Magic. This package is only of use to current PRB members and will require you to make a concrete5.org login. Your username and password are passed through to concrete5.org and not held on the site.');
    }

    public function getPackageName() {
        return t('Package Magic PRB');
    }

    public function validate_install() {
        $pkg_m = Package::getByHandle('jl_package_magic_starter');
        if (!is_object($pkg_m)) {
            $err = Application::getFacadeApplication()->make('error');
            $err->add(t('Package Magic Starter must be installed before Package Magic PRB. This package provides an extension for Package Magic.'));
            return $err;
        }
    }

    public function install() {
        $pkg = parent::install();
        $this->do_install($pkg);
    }

    public function upgrade() {
        parent::upgrade();
        $this->do_install($this);
    }

    private function do_install($pkg) {
        /*
         * Sort out resources and Re-scan for Package Magic Source plugins
         */
        $app = Application::getFacadeApplication();
        $app->make('JtF\PackageMagic\PackageSources\PackageSourcesController')->listAll(true);
    }

    /*
     * Remove any cached PRB list or concrete5.org cookie details
     */

    public function uninstall() {
        $this->logout_concrete5();
        parent::uninstall();
        $app = Application::getFacadeApplication();
        $app->make('JtF\PackageMagic\PackageSources\PackageSourcesController')->listAll(true);
    }

    public function getPackageAutoloaderRegistries() {
        $auto = array(
            'src/JtF/PackageMagic' => '\\JtF\\PackageMagic'
        );
        return $auto;
    }

    /*
     * Flush cookie cache file when cache is flushed
     */

    public function on_start() {
        Events::addListener('on_cache_flush', function () {
            $this->logout_concrete5();
        }
        );
    }

    /*
     * Clear all cookies and records, so forcing a logout
     */

    private function logout_concrete5() {
        $app = Application::getFacadeApplication();
        $cache = $app->make('cache/expensive');
        $cache->delete('jl_package_magic_from_prb_fetch_PRB_list_page');
        $fs = new Filesystem();
        $fs->delete(DIR_CONFIG_SITE . '/jl_package_magic_concrete5_login_cookie.txt');
    }

}
