<?php

namespace JtF\PackageMagic\PackageSources\Plugins;

use Page;
use JtF\PackageMagic\PackageSources\PackageSourcePluginBase;
use JtF\PackageMagic\PackageZipValidation;

defined('C5_EXECUTE') or die('Access Denied');
/*
 * Package Magic by John Liddiard (aka JohntheFish)
 * www.c5magic.co.uk
 *
 * This package contains software copyright and proprietary to John Liddiard
 */

class FromPRB extends PackageSourcePluginBase {

    protected $curl_timeout = 20;

    public function getName() {
        return t('From PRB');
    }

    public function getDescription() {
        return t('Pick a submission from the PRB.');
    }

    public function getHelp() {
        return t('This source is only relevant to PRB members. List PRB submissions and install them directly to your PRB test site. You will be required to provide your concrete5.org login and password, but these will not be saved.');
    }

    /*
     * Fragment for the edit form dialog
     *
     *
     * https://stackoverflow.com/questions/20049393/using-php-curl-to-login-to-my-websites-form
     * http://blog.andyhunt.info/2011/12/21/using-php-and-curl-to-log-in-to-a-website/
     * https://stackoverflow.com/questions/3008817/login-to-remote-site-with-php-curl
     * https://www.sitepoint.com/using-curl-for-remote-requests/
     *
     */

    public function edit($controller, $data) {
        $data = $this->synthesizeData($data);
        extract($controller->getSets());
        $form = $this->form();

        ob_start();

        if ($this->is_force_login()) {
            self::logout_concrete5();
            $this->clear_force_login();
            $submissions = false;
        } else {
            /*
             * TODO. For the future, either an option to show submissions
             * that have failed auto test, or merge the failed test list and
             * highlight status (red/green)
             */
            $submissions = $this->fetch_all_PRB_lists($controller, $data);
        }
        ?>
        <style>

            .from_prb_wrapper{
                margin-top:10px;
            }
            .prb_hover_help{
                cursor:help;
            }
        </style>
        <div class="from_prb_wrapper">
            <?php
            if ($submissions === false) {
                ?>
                <div style="display:block;opacity:0;height:1px;overflow:hidden;">
                    <?php
                    /*
                     * Trick to prevent browsers from filling in the actual fields below. They unhelpfully use the
                     * site user/pw and we don't want those at c5.org
                     */
                    echo $form->label('DummyFooleBrowserMemoryuName', t('Dummy Username'));
                    echo $form->text('DummyFooleBrowserMemoryuName', '', array('placeholder' => t(t('Username'))));
                    echo $form->label('DummyFooleBrowserMemoryuPassword', t('Dummy Password'));
                    echo $form->password('DummyFooleBrowserMemoryuPassword', '', array('placeholder' => t(t('Password'))));
                    ?>
                </div>
                <legend>
                    <?php
                    echo t('Sign in to concrete5.org');
                    ?>
                </legend>

                <div class="fieldset" style="margin-bottom:1000px;">

                    <div class="form-group">
                        <?php
                        echo $form->label('c5uName', t('Username'));
                        echo $form->text('c5uName', '', array('placeholder' => t(t('Username'))));
                        ?>
                    </div>

                    <div class="form-group">
                        <?php
                        echo $form->label('c5uPassword', t('Password'));
                        echo $form->password('c5uPassword', '', array('placeholder' => t(t('Password'))));
                        ?>
                    </div>
                    <div class="form-group">
                        <?php
                        echo $form->submit('c5_submit', t('Login'), array('class' => 'btn btn-primary pull-right'));
                        ?>
                    </div>
                </div>
                <?php
            } else {
                $this->clear_force_login();
                if (!empty($submissions)) {
                    /*
                     * TODO, sort/filter list. Maybe implement using same list mechanism as
                     * PM plugins, so also searchable.
                     */
                    ?>
                    <table class="table table-striped">
                        <thead>
                            <tr>
                                <th>

                                </th>
                                <th>

                                </th>
                                <th>
                                    <?php
                                    echo t('Item')
                                    ?>
                                </th>
                                <th>
                                    <?php
                                    echo t('Linter')
                                    ?>
                                </th>
                                <th>
                                    <?php
                                    echo t('Author')
                                    ?>
                                </th>
                                <th>
                                    <?php
                                    echo t('Added on')
                                    ?>
                                </th>
                                <th>
                                    <?php
                                    echo t('Last comment')
                                    ?>
                                </th>
                                <th>
                                    <?php
                                    echo t('Will go live')
                                    ?>
                                </th>
                                <th>
                                    <?php
                                    echo t('Last reply')
                                    ?>
                                </th>
                            </tr>
                        </thead>
                        <tbody>
                            <?php
                            /*
                             * TODO, CHECK IF INSTALLED
                             * Something for the future. Not as simple as it may sound as
                             * submissions have names, but the list doesn't show handles,
                             * so could fail when submission names change.
                             */
                            foreach ($submissions as $submission) {
                                ?>
                                <tr data_page="<?php echo h($submission['submission_id']); ?>">
                                    <td>
                                        <?php
                                        echo $form->radio($this->prefixNamespace('submission_id'), h($submission['submission_id']));
                                        ?>
                                    </td>
                                    <td>
                                        <?php
                                        if ($submission['type'] === 'Addon' || $submission['type'] === t('Addon')) {
                                            ?>
                                            <i class="fa fa-cog"></i>
                                            <?php
                                        } else {
                                            ?>
                                            <i class="fa fa-television"></i>
                                            <?php
                                        }
                                        ?>
                                    </td>
                                    <td>
                                        <a href="<?php echo sprintf(self::REVIEW_PAGE, $submission['submission_id']); ?>" target="_blank" title="<?php echo t('Review thread on concrete5.org'); ?>" data-toggle="title">
                                            <?php
                                            echo $submission['name'];
                                            ?>
                                        </a>
                                    </td>
                                    <td>
                                        <?php
                                        if ($submission['linter_status'] === 'failed') {
                                            ?>
                                            <i class="text-danger fa fa-times-circle prb_hover_help" title="<?php echo t('Failed automated tests') ?>"></i>
                                            <?php
                                        } else if ($submission['linter_status'] === 'testing') {
                                            ?>
                                            <i class="text-muted fa fa-question-circle prb_hover_help" title="<?php echo t('Tests in progress') ?>"></i>
                                            <?php
                                        } else {
                                            ?>
                                            <i class="text-success fa fa-check-circle prb_hover_help" title="<?php echo t('Passed automated tests') ?>"></i>
                                            <?php
                                        }
                                        ?>
                                    </td>
                                    <td>
                                        <?php
                                        echo $submission['developer_name'];
                                        ?>
                                    </td>
                                    <td>
                                        <?php
                                        echo $submission['submission_date']
                                        ?>
                                    </td>
                                    <td>
                                        <?php
                                        echo $submission['comment_date']
                                        ?>
                                    </td>
                                    <td>
                                        <?php
                                        echo $submission['time_remaining']
                                        ?>
                                    </td>
                                    <td>
                                        <?php
                                        echo $submission['comment_by']
                                        ?>
                                    </td>
                                </tr>
                                <?php
                            }
                            ?>
                        </tbody>
                    </table>
                    <?php
                } else {
                    ?>
                    <div class="alert alert-danger">
                        <?php
                        echo t('The PRB list is empty!');
                        ?>
                    </div>
                    <?php
                }
                ?>
                <div class="form-group">
                    <a href="#c5_logout" class="btn btn-default pull-left"><i class="fa fa-close"> </i>
                        <?php
                        echo t('Logout from concrete5.org');
                        ?>
                    </a>
                    <a href="#c5_prb_refresh" class="btn btn-info pull-right"><i class="fa fa-refresh"> </i>
                        <?php
                        echo t('Refresh');
                        ?>
                    </a>
                </div>

                <?php
            }
            ?>
        </div>
        <?php
        $page = Page::getCurrentPage();
        $cID = $page->getCollectionID();
        $args = array(
            'ccm_token' => $this->make_token(),
            'called_class' => __CLASS__,
            'action' => 'from_prb_refresh_ajax',
            'cID' => $cID
        );
        $args['sk'] = $this->encrypt($args);
        $url = self::app()->make('helper/url')->buildQuery($this->ajax_controller_url(), $args);
        $args2 = array(
            'ccm_token' => $this->make_token(),
            'called_class' => __CLASS__,
            'action' => 'from_prb_logout_ajax',
            'cID' => $cID
        );
        $args2['sk'] = $this->encrypt($args2);
        $url2 = self::app()->make('helper/url')->buildQuery($this->ajax_controller_url(), $args2);
        ?>
        <script type="text/javascript">
            (function ($) {
                $(document).ready(function () {
                    var this_form = $('#package_magic_upload');

                    this_form.find('a[href="#c5_prb_refresh"]').one('click', function (ev) {
                        ev.preventDefault();
                        $.fn.dialog.showLoader();
                        $.ajax({
                            url: '<?php echo $url; ?>',
                            type: 'get',
                        }).always(function () {
                            $.fn.dialog.hideLoader();
                            window.location.reload();
                        });
                    });

                    this_form.find('a[href="#c5_logout"]').one('click', function (ev) {
                        ev.preventDefault();
                        $.fn.dialog.showLoader();
                        $.ajax({
                            url: '<?php echo $url2; ?>',
                            type: 'get',
                        }).always(function () {
                            $.fn.dialog.hideLoader();
                            window.location.reload();
                        });
                    });
                });
            })(jQuery);
        </script>
        <?php
        $opts_HTML = ob_get_contents();
        ob_end_clean();
        return $opts_HTML;
    }

    public function from_prb_refresh_ajax($controller, $data) {
        $cache = self::app()->make('cache/expensive');
        $cache->delete(self::CACHE_KEY);
        $cache->delete(self::CACHE_KEY . 'testing');
        $cache->delete(self::CACHE_KEY . 'failed');
        $ajax = self::app()->make('helper/ajax')->sendResult(array('msg' => 'refresh'));
        exit;
    }

    public function from_prb_logout_ajax($controller, $data) {
        self::logout_concrete5();
        $this->set_force_login();
        self::app()->make('helper/ajax')->sendResult(array('msg' => 'logout'));
        exit;
    }

    /*
     * Gets all prb lists, regular, testing and failed.
     */
    private function fetch_all_PRB_lists($controller, $data) {
        $submissions = $this->fetch_PRB_list($controller, $data);
        $submissions_testing = $this->fetch_PRB_list($controller, $data, 'testing');
        $submissions_failed = $this->fetch_PRB_list($controller, $data, 'failed');
        $submissions = array_merge($submissions, $submissions_testing, $submissions_failed);
        return $submissions;
    }

    /*
     * Gets the (cached) PRB list and scrapes it for submissions
     *
     * Any more complex and this would be easier with phpquery or even pass this on to the
     * view and use jQuery. However, the page is pretty much static in design and provides a
     * table, so can get away with quick and dirty regexes and strstr
     */

    private function fetch_PRB_list($controller, $data, $status = null) {
        $page = $this->fetch_PRB_list_cached_page($controller, $data, $status);
        if ($page !== false) {

            $p1 = strstr($page, 'id="prb"');
            $p1 = strstr($p1, '/thead>');
            $p1 = strstr($p1, '<tr>');
            $p1 = strstr($p1, '</table>', true);
            $rows = array_filter(explode('<tr>', $p1));
            if (empty($rows)) {
                return array();
            }

            $submission_rows = array();
            foreach ($rows as $row) {
                $cols = explode('<td>', $row);
                $info = array();

                if (strpos($cols[0], 'prb_add_on') !== false) {
                    $info['type'] = t('Addon');
                } else {
                    $info['type'] = t('Theme');
                }

                if (preg_match("/href=\".*?\/([0-9]*)\/\">(.*?)</", $cols[0], $m)) {
                    $info['submission_id'] = $m[1];
                    $info['name'] = $m[2];
                }

                if (preg_match("/href=\"(.*?)\">(.*?)</", $cols[1], $m)) {
                    $info['developer_page'] = $m[1];
                    $info['developer_name'] = $m[2];
                }

                $info['submission_date'] = trim(strip_tags($cols[2]));
                $info['comment_date'] = trim(strip_tags($cols[3]));
                $info['time_remaining'] = trim(strip_tags($cols[4]));
                $info['comment_by'] = trim(strip_tags($cols[5]));
                $info['linter_status'] = $status;

                $submission_rows[$info['submission_id']] = $info;
            }
            if (!empty($submission_rows)) {
                return $submission_rows;
            }
            return array();
        }
        return array();
    }

    const CACHE_KEY = 'jl_package_magic_from_prb_fetch_PRB_list_page';
    const REVIEW_LIST = "https://www.concrete5.org/marketplace/review_pending/";
    const DO_LOGIN = "https://www.concrete5.org/login/-/do_login/";
    const LOGIN = "https://www.concrete5.org/login/";
    const DOWNLOAD_PACKAGE = 'https://www.concrete5.org/marketplace/review_pending/-/download/%d/1/';
    const REVIEW_PAGE = 'https://www.concrete5.org/marketplace/review_pending/-/view_detail/%d/';

    // const COOKIE_FILE = DIR_CONFIG_SITE . '/jl_package_magic_concrete5_login_cookie.txt'; // PHP version compatibility
    private function cookie_file() {
        return DIR_CONFIG_SITE . '/jl_package_magic_concrete5_login_cookie.txt';
    }

    /*
     * Using the core expensive cache. May develop my own tabled cache later.
     * A bit complex because it has to adapt (or at least not fail) when the
     * core api is continually farting about.
     */

    private function fetch_PRB_list_cached_page($controller, $data, $status = null) {
        $cache = self::app()->make('cache/expensive');
        $r = $cache->getItem(self::CACHE_KEY . $status);
        if ((method_exists($r, 'isHit') && $r->isHit()) ||
                (method_exists($r, 'isMiss') && !$r->isMiss())
        ) {
            if (method_exists($r, 'get')) {
                $page = $r->get();
                if (!empty($page)) {
                    return $page;
                }
            }
        }
        $page = $this->fetch_PRB_list_page($controller, $data, $status);

        if (method_exists($r, 'set')) {
            if (method_exists($cache, 'save')) {
                $cache->save($r->set($page));
            } else {
                $r->set($page);
            }
        }
        return $page;
    }

    private function fetch_PRB_list_page($controller, $data, $status = null) {

        $post_data = $controller->post();
        try {
            if ($post_data['c5uName'] && $post_data['c5uPassword']) {
                $params = array(
                    'uName' => $post_data['c5uName'],
                    'uPassword' => $post_data['c5uPassword'],
                    'uMaintainLogin' => 1, // always maintain login, or we loose continuity between requests
                    'rcID' => "",
                    'ccm_token' => $this->get_c5_login_token(),
                );
                $this->fetch_page(self::DO_LOGIN, $params);
            }
            if ($status) {
                $page = $this->fetch_page(self::REVIEW_LIST . '?status=' . $status);
            } else {
                $page = $this->fetch_page(self::REVIEW_LIST);
            }
            /*
             * Simple quick-test we have a prb list and not a login
             * page again
             */
            if (strpos($page, 'id="prb"') !== false) {
                return $page;
            }
        } catch (\Exception $e) {
            // its failed, so return false;
        }
        return false;
    }

    private function get_c5_login_token() {
        $page = $this->fetch_page(self::LOGIN);
        if (preg_match("/name=\"ccm_token\".*?value=\"(.*?)\"/s", $page, $m)) {
            return $m[1];
        }
    }

    /*
     * Fragment for the edit form dialog
     *
     *
     * https://stackoverflow.com/questions/20049393/using-php-curl-to-login-to-my-websites-form
     * http://blog.andyhunt.info/2011/12/21/using-php-and-curl-to-log-in-to-a-website/
     * https://stackoverflow.com/questions/3008817/login-to-remote-site-with-php-curl
     * https://www.sitepoint.com/using-curl-for-remote-requests/
     *
     */

    private function fetch_page($url, $params = null) {

        $ch = curl_init();
        curl_setopt($ch, CURLOPT_URL, $url);

        if (!empty($params) && is_array($params)) {
            curl_setopt($ch, CURLOPT_POST, 1);
            curl_setopt($ch, CURLOPT_POSTFIELDS, $params);
        }
        curl_setopt($ch, CURLOPT_TIMEOUT, $this->curl_timeout);
        /*
         * Force curl to return rather than output
         */
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
        curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, false);
        /*
         * Follow redirects
         */
        curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);

        /*
         * Cookies from & to file
         */
        curl_setopt($ch, CURLOPT_COOKIEFILE, $this->cookie_file());
        curl_setopt($ch, CURLOPT_COOKIEJAR, $this->cookie_file());
        $response = curl_exec($ch);
        $error = curl_error($ch);
        curl_close($ch);
        if (!empty($error)) {
            \Log::addError(t('CURL error fetching page %s, %s', $url, $error));
            $this->puke(t('CURL error fetching page %s, %s', $url, $error));
        }

        return $response;
    }

    /*
     * Header needed to get filename and check we have a zip
     * https://stackoverflow.com/questions/9183178/can-php-curl-retrieve-response-headers-and-body-in-a-single-request
     */

    private function fetch_download($id, $status = null) {
        $url = sprintf(self::DOWNLOAD_PACKAGE, $id);
        $headers = array();

        $ch = curl_init();
        curl_setopt($ch, CURLOPT_URL, $url);

        curl_setopt($ch, CURLOPT_TIMEOUT, $this->curl_timeout);
        /*
         * Force curl to return rather than output
         */
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_HEADERFUNCTION, function($ch, $header) use (&$headers) {
            /*
             * $headers is updated and normaised.
             */
            $len = strlen($header);
            $header = explode(':', $header, 2);
            if (count($header) < 2) // ignore invalid headers
                return $len;

            $name = strtolower(trim($header[0]));
            if (!array_key_exists($name, $headers)) {
                $headers[$name] = [trim($header[1])];
            } else {
                $headers[$name][] = trim($header[1]);
            }
            return $len;
        }
        );
        curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
        curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, false);

        /*
         * Cookies from & to file
         */
        curl_setopt($ch, CURLOPT_COOKIEFILE, $this->cookie_file());
        curl_setopt($ch, CURLOPT_COOKIEJAR, $this->cookie_file());

        $response = curl_exec($ch);
        $error = curl_error($ch);
        curl_close($ch);
        if (!empty($error)) {
           \Log::addError(t('CURL error fetching download %s, %s', $url, $error));
            $this->puke(t('CURL error fetching download %s, %s', $url, $error));
        }

        $header_err = $this->check_headers($headers);

        if (!empty($header_err)) {
            \Log::addError(t('Header error fetching download %s, %s', $url, $header_err));
            $this->puke(t('Header error fetching download %s, %s', $url, $header_err));
        }
        return $response;
    }

    private $submission_id;
    private $submission_package_handle;

    private function check_headers($headers) {
        if (!empty($headers['content-disposition']) && is_array($headers['content-disposition'])) {
            $content_disposition = reset(array_filter($headers['content-disposition']));
            if (!empty($content_disposition)) {
                if (preg_match("/filename=\"(.*?)\.zip\"/", $content_disposition, $m)) {
                    $this->submission_package_handle = $m[1];
                    return;
                }
            }
        }
        /*
         * IF the above fails, log the headers for future diagnosis.
         */
        $this->log('Error', json_encode($headers, JSON_PRETTY_PRINT));
        return t('Download failed, you may need to logout and login again to get a clean connection.');
    }

    /*
     * After selecting a package, configure the plugin with a package it will
     * provide to the client. As the file manager may be storing the file remotely,
     * we get the content and save a temporary copy that can be deleted on cleanup.
     *
     * https://documentation.concrete5.org/developers/working-with-files-and-the-file-manager/accessing-the-physical-file-from-a-file-object
     *
     */

    public function initialiseSource($controller, $data) {
        $data = $this->synthesizeData($data);

        $submissions = $this->fetch_all_PRB_lists($controller, $data);

        extract($controller->getSets());
        $this->submission_id = $data['submission_id'];

        if (empty($this->submission_id)) {
            $this->puke(t('Please select a package from the PRB list'));
        }

        $submission = $submissions[$data['submission_id']];
        if (empty($this->submission_id)) {
            $cache = self::app()->make('cache/expensive');
            $cache->delete(self::CACHE_KEY);
            $cache->delete(self::CACHE_KEY . 'testing');
            $cache->delete(self::CACHE_KEY . 'failed');
            $this->puke(t('Submission data is obsolete, please try again'));
        }

        $this->intermediate_file = $this->file()->getTemporaryDirectory() . DIRECTORY_SEPARATOR .
                $this->getClassName(__CLASS__) . '-intermediate-' . $this->submission_id . '.zip';
        if (file_exists($this->intermediate_file)) {
            $this->file()->delete($this->intermediate_file);
        }
        if (file_exists($this->intermediate_file)) {
            $this->puke(t('Unable to remove previous intermediate file %s', $this->intermediate_file));
        }

        /*
         * Get the file, from the 'package' url. Do not trim as this will corrupt the zip!!
         */
        $remote_content = $this->fetch_download($this->submission_id);

        if (empty($remote_content)) {
            $cache = self::app()->make('cache/expensive');
            $cache->delete(self::CACHE_KEY);
            $cache->delete(self::CACHE_KEY . 'testing');
            $cache->delete(self::CACHE_KEY . 'failed');
            $this->puke(t('Unable to retrieve submission %s', $submission['name']));
        }
        $this->file()->append($this->intermediate_file, $remote_content);
    }

    /*
     * No cleanup
     */

    /*
     * Validation specific to this source
     */

    public function validate() {
        if (empty($this->submission_id)) {
            return t('Please select a PRB submission.');
        }
        if (empty($this->intermediate_file) || !file_exists($this->intermediate_file)) {
            return t('Temporary intermediate file %s is not accessible.', $this->intermediate_file);
        }

        if (filesize($this->intermediate_file) <= 0) {
            return t('File "%s" is empty', $this->submission_id);
        }
        $zip_val = new PackageZipValidation($this->intermediate_file);
        if (!$zip_val->is_valid_mime()) {
            return t('Source is not a zip archive');
        }
        return;
    }

    /*
     * A path to the zip file containing the package
     */

    public function getZipPath() {
        return $this->intermediate_file;
    }

    /*
     * The package handle
     */

    public function getPackageHandle() {
        if (!empty($this->submission_package_handle)) {
            return $this->submission_package_handle;
        }
        return $this->cleanHandle(basename($this->submission_id));
    }

    /*
     * Login action flag passed between requests by session
     */

    private function is_force_login() {
        $session = self::app()->make('session');
        $force_login = $session->get('jl_package_magic.upload.from_prb_logout');
        return !empty($force_login);
    }

    private function set_force_login() {
        $session = self::app()->make('session');
        $session->set('jl_package_magic.upload.from_prb_logout', true);
    }

    private function clear_force_login() {
        $session = self::app()->make('session');
        $session->set('jl_package_magic.upload.from_prb_logout', false);
    }

    /*
     * Clear all cookies and records, so forcing a logout
     */

    public static function logout_concrete5() {
        $cache = self::app()->make('cache/expensive');
        $cache->delete('jl_package_magic_from_prb_fetch_PRB_list_page');
        self::file_s()->delete(DIR_CONFIG_SITE . '/jl_package_magic_concrete5_login_cookie.txt');
    }

}
